import os
import numpy as np
from pysad.utils.array_streamer import ArrayStreamer


class Data:
    """A helper class to load various data.

    Args:
        data_base_path (str): Base path that contains the data files.
    """

    def __init__(self, data_base_path="data"):
        self.data_base_path = data_base_path

    def _get_data_files(self):
        """ Helper method to return the names of the data files.

        Returns:
            file_names (list[str]): List of data file names.
        """
        return [
            'arrhythmia.mat',
            'cardio.mat',
            'glass.mat',
            'ionosphere.mat',
            'letter.mat',
            'lympho.mat',
            'mnist.mat',
            'musk.mat',
            'optdigits.mat',
            'pendigits.mat',
            'pima.mat',
            'satellite.mat',
            'satimage-2.mat',
            'shuttle.mat',
            'vertebral.mat',
            'vowels.mat',
            'wbc.mat',
            "gisette_sampled.txt",
            "isolet_sampled.txt",
            "madelon_sampled.txt",
            "pima-indians_sampled.txt",
            "magic-telescope_sampled.txt",
        ]

    def _load_via_txt(self, path):
        """Loads the data file from .txt file.

        Args:
            path (str): The path of data.

        Returns:
            X (np.float64 array of shape (num_instances, num_features)): Feature vectors.
        """
        X = np.loadtxt(path, delimiter=",")

        return X

    def get_data(self, data_file):
        """Loads the data given the path.

        Args:
            data_file: Path of the data.

        Returns:
            X (np.array of shape (num_instances, num_features)): Feature vectors.
            y (np.array of shape (num_instances,)): Labels.
        """
        data_path = os.path.join(self.data_base_path, data_file)

        if ".mat" in data_file:
            from scipy.io import loadmat

            f = loadmat(data_path)

            X = f['X']
            y = f['y'].ravel()
        else:
            X = self._load_via_txt(data_path)

            y = X[:, -1].ravel()
            X = X[:, :-1]

        return X, y

    def get_iterator(self, data_file, shuffle=True, seed=None):
        """The iterator function

        Args:
            data_file (str): Path of data.
            shuffle (bool): Whether to shuffle (Default=True).
            seed (int): Random seed (Default=None).

        Returns:
            iterator (The iterator): pysad.utils.array_streamer.ArrayStreamer.iter method applied with (X, y), where X is the variable containing feature vectors and y is the variable containing labels.

        """
        if seed is not None:
            np.random.seed(seed)

        iterator = ArrayStreamer(shuffle=shuffle)

        X, y = self.get_data(data_file)

        return iterator.iter(X, y)



""""
The :mod:`pysad.transform.preprocessing` module includes preprocessing methods to transform inputs such as normalizers.
"""
from .identity_scaler import IdentityScaler
from .instance_standard_scaler import InstanceStandardScaler
from .instance_unit_norm_scaler import InstanceUnitNormScaler

__all__ = ["IdentityScaler", "InstanceStandardScaler", "InstanceUnitNormScaler"]



def helper_test_all_metrics(metric_classes, y_true, y_pred, ignore_nonempty_last):
    import numpy as np
    from pysad.evaluation import WindowedMetric

    for metric_cls, val in metric_classes.items():
        metric = WindowedMetric(metric_cls, 25, ignore_nonempty_last)

        for i, (yt, yp) in enumerate(zip(y_true, y_pred)):
            metric.update(yt, yp)
            if i > 0:
                assert np.isclose(metric.get(), val)


def test_all_correct():
    from pysad.evaluation import PrecisionMetric, AUPRMetric, AUROCMetric, RecallMetric
    import numpy as np
    metric_classes = [
        PrecisionMetric,
        RecallMetric,
        AUPRMetric,
        AUROCMetric
    ]
    metric_classes = { metric_cls: 1.0 for metric_cls in metric_classes }
    y_true = np.random.randint(0, 2, size=(25,), dtype=np.int32)
    y_true[0] = 1
    y_true[1] = 0
    y_pred = y_true.copy()

    for ignore_nonempty_last in [True, False]:
        helper_test_all_metrics(metric_classes, y_true, y_pred, ignore_nonempty_last)


def test_none_correct():
    from pysad.evaluation import PrecisionMetric, AUPRMetric, AUROCMetric, RecallMetric
    import numpy as np

    metric_classes = {
        PrecisionMetric: 0.0,
        #AUPRMetric: 0.5
        AUROCMetric: 0.0,
        RecallMetric: 0.0
    }
    y_true = np.random.randint(0, 2, size=(25,), dtype=np.int32)
    y_true[0] = 1
    y_true[1] = 0
    y_pred = 1 - y_true.copy()

    for ignore_nonempty_last in [True, False]:
        helper_test_all_metrics(metric_classes, y_true, y_pred, ignore_nonempty_last)


def test_windowed_metric_window_size():
    """Test WindowedMetric with different window sizes."""
    from pysad.evaluation import WindowedMetric, AUROCMetric
    
    # Test small window
    metric = WindowedMetric(AUROCMetric, window_size=3)
    
    # Add data beyond window size
    for i in range(5):
        y_true = i % 2
        y_pred = 0.5 + (i % 2) * 0.3
        metric.update(y_true, y_pred)
    
    score = metric.get()
    assert isinstance(score, (int, float))


def test_windowed_metric_ignore_nonempty_last():
    """Test WindowedMetric ignore_nonempty_last parameter."""
    from pysad.evaluation import WindowedMetric, AUROCMetric  # Use AUROC which accepts continuous scores
    
    # Test with ignore_nonempty_last=True
    metric_ignore = WindowedMetric(AUROCMetric, window_size=5, ignore_nonempty_last=True)
    
    # Test with ignore_nonempty_last=False  
    metric_no_ignore = WindowedMetric(AUROCMetric, window_size=5, ignore_nonempty_last=False)
    
    # Add same data to both with mixed classes and continuous scores
    test_data = [(1, 0.8), (0, 0.2), (1, 0.9)]
    for y_true, y_pred in test_data:
        metric_ignore.update(y_true, y_pred)
        metric_no_ignore.update(y_true, y_pred)
    
    # Both should work but may have different behavior
    score_ignore = metric_ignore.get()
    score_no_ignore = metric_no_ignore.get()
    
    assert isinstance(score_ignore, (int, float))
    assert isinstance(score_no_ignore, (int, float))


def test_windowed_metric_empty_window():
    """Test WindowedMetric behavior with empty window."""
    from pysad.evaluation import WindowedMetric, RecallMetric
    import warnings
    
    metric = WindowedMetric(RecallMetric, window_size=10)
    
    # Get score before any updates
    # Suppress the sklearn warning since we're testing edge cases  
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Recall is ill-defined and being set to 0.0")
        score = metric.get()
        assert score == 0.0


def test_windowed_metric_various_metrics():
    """Test WindowedMetric with different base metrics."""
    from pysad.evaluation import WindowedMetric, AUPRMetric, AUROCMetric, PrecisionMetric, RecallMetric
    import numpy as np
    
    # Separate metrics by data type they expect
    continuous_metrics = [
        WindowedMetric(AUPRMetric, window_size=10),
        WindowedMetric(AUROCMetric, window_size=10)
    ]
    
    binary_metrics = [
        WindowedMetric(PrecisionMetric, window_size=10),
        WindowedMetric(RecallMetric, window_size=10)
    ]
    
    # Add diverse data with continuous scores
    np.random.seed(42)
    for _ in range(15):
        y_true = np.random.randint(0, 2)
        y_pred_continuous = np.random.random()
        y_pred_binary = np.random.randint(0, 2)
        
        # Test continuous metrics with continuous predictions
        for metric in continuous_metrics:
            metric.update(y_true, y_pred_continuous)
            
        # Test binary metrics with binary predictions
        for metric in binary_metrics:
            metric.update(y_true, y_pred_binary)
    
    # All metrics should return valid scores
    for metric in continuous_metrics + binary_metrics:
        score = metric.get()
        assert isinstance(score, (int, float))
        assert 0.0 <= score <= 1.0



from abc import abstractmethod, ABC


class BaseStatistic(ABC):
    """Abstact base class for the statistics.
    """
    pass


class UnivariateStatistic(BaseStatistic):
    """Abstract base class for univariate statistics.
    """

    @abstractmethod
    def update(self, num):
        """Updates the statistic with the value for a timestep.

        Args:
            num (float):  The incoming value, for which the statistic is used.

        Returns:
            object: self.
        """
        pass

    @abstractmethod
    def get(self):
        """ Method to obtain the tracked statistic.

        Returns:
            float: The statistic.
        """
        pass

    @abstractmethod
    def remove(self, num):
        """Updates the statistic by removing particular value. This method

        Args:
            num (float): The value to be removed.

        Returns:
            object: self.
        """
        pass



from scipy import stats
from pysad.core.base_model import BaseModel
import math
import numpy as np


class RelativeEntropy(BaseModel):
    """Relative entropy based anomaly detection model on univariate stream :cite:`ahmad2017unsupervised`. The implementation is based on `NAB-relative_entropy <https://github.com/numenta/NAB/blob/master/nab/detectors/relative_entropy/relative_entropy_detector.py>`_.

        Args:
            min_val (float): Minimum value of the univariate stream.
            max_val (float): Maximum value of the univariate stream.
            num_bins (int): Number of bins (Default=5).
            window_size (int): The size of the window (Default=52).
    """

    def __init__(self, min_val, max_val, num_bins=5, window_size=52):
        self.min_val = min_val
        self.max_val = max_val

        # Timeseries of the metric on which anomaly needs to be detected
        self.util = []

        # Number of bins into which util is to be quantized
        self.N_bins = num_bins

        # Window size
        self.W = window_size

        # Threshold against which the test statistic is compared. It is set to
        # the point in the chi-squared cdf with N-bins -1 degrees of freedom that
        #  corresponds to 0.99.
        self.T = stats.chi2.isf(0.01, self.N_bins - 1)

        # Tracks the current number of null hypothesis
        self.m = 0

        # Step size in time series quantization
        self.stepSize = (max_val - min_val) / self.N_bins

        # List of lists where P[i] indicates the empirical frequency of the ith
        # hypothesis.
        self.P = []

        # List where c[i] tracks the number of windows that agree with P[i]
        self.c = []

        self.P_hat = None

    def fit_partial(self, X, y=None):
        """Fits the model to next instance.

        Args:
            X (float): The instance to fit. Note that this model is univariate.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: Returns the self.
        """
        self.util.append(X)
        if len(self.util) >= self.W:

            # Extracting current window
            util_current = self.util[-self.W:]

            # Quantize window data points into discretized bin values
            B_current = [math.ceil((np.asarray(c).item() - self.min_val) / self.stepSize) for c in
                         util_current]

            # Create a histogram of empirical frequencies for the current window
            # using B_current
            self.P_hat = np.histogram(B_current,
                                      bins=self.N_bins,
                                      range=(0, self.N_bins),
                                      density=True)[0]

            if self.m == 0:
                self.P.append(self.P_hat)
                self.c.append(1)
                self.m = 1

        return self

    def score_partial(self, X):
        """Scores the anomalousness of the next instance. Note that this method should be called after the fit_partial method.

        Args:
            X (any): (Ignored) The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            float: The anomalousness score of the input instance.
        """
        score = 0.0

        if len(self.util) >= self.W and self.m > 0 and self.P_hat is not None:
            score = self._get_aggreement_hypothesis(self.P_hat)

        return score

    def _get_aggreement_hypothesis(self, P_hat):
        """This function computes multinomial goodness-of-fit test. It calculates
        the relative entropy test statistic between P_hat and all `m` null
        hypothesis and compares it against the threshold `T` based on cdf of
        chi-squared distribution. The test relies on the observation that if the
        null hypothesis P is true, then as the number of samples grow the relative
        entropy converges to a chi-squared distribution1 with K-1 degrees of
        freedom.
        The function returns the index of hypothesis that agrees with minimum
        relative entropy. If all hypotheses disagree, the function returns -1.
        @param P_hat    (list)  Empirical frequencies of the current window.
        @return index   (int)   Index of the hypothesis with the minimum test
                                statistic.
        """

        index = -1
        minEntropy = float("inf")
        for i in range(self.m):
            entropy = 2 * self.W * stats.entropy(P_hat, self.P[i])
            if entropy < self.T and entropy < minEntropy:
                minEntropy = entropy
                index = i

        return index



from .pyod_model import PYODModel
import numpy as np


class ReferenceWindowModel(PYODModel):
    """This PyOD model wrapper wraps the batch anomaly detectors. This wrapper keeps track of the reference window of size `window_length`. For every `sliding_size` instnaces, it resets the model by training new `model_cls` instance with the reference window. This implementation is based on the reference windowing described in :cite:`xstream`.

        Args:
            model_cls (class): The model class to be instantiated.
            window_size (int): The size of each window.
            sliding_size (int): The sliding length of the windows.
            initial_X (np.float64 array of shape (num_initial_instances, num_features)): Initial instances to fit.
            initial_y (np.int32 array of shape (num_initial_instances,)): Initial window's ground truth labels. Used if not None. Needs to be `None` for the unsupervised `model_cls` models. (Default=None).
            **kwargs (Keyword arguments): Keyword arguments that is passed to the `model_cls`.
    """

    def __init__(
            self,
            model_cls,
            window_size,
            sliding_size,
            initial_window_X=None,
            initial_window_y=None,
            **kwargs):
        """

        Args:
            model_cls:
            window_size:
            sliding_size:
            initial_window_X:
            initial_window_y:
        """
        super().__init__(model_cls, **kwargs)

        self.window_size = window_size
        self.sliding_size = sliding_size

        self.cur_window_X = []
        self.cur_window_y = []

        self.reference_window_X = initial_window_X
        self.reference_window_y = initial_window_y

        if self.reference_window_X is not None:
            self._fit_model()
            self.initial_ref_window = True
        else:
            self.initial_ref_window = False

    def fit_partial(self, X, y=None):
        """Fits the model to next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to fit.
            y (int): The label of the instance (Optional for unsupervised models, default=None).

        Returns:
            object: self.
        """
        self.cur_window_X.append(X)

        if y is not None:
            self.cur_window_y.append(y)

        if not self.initial_ref_window and len(
                self.cur_window_X) < self.window_size:
            self.reference_window_X = self.cur_window_X
            self.reference_window_y = self.cur_window_y if y is not None else None
            self._fit_model()
        elif len(self.cur_window_X) % self.sliding_size == 0:
            self.reference_window_X = np.concatenate(
                [self.reference_window_X, self.cur_window_X], axis=0)
            self.reference_window_X = self.reference_window_X[max(
                0, len(self.reference_window_X) - self.window_size):]

            if y is not None:
                self.reference_window_y = np.concatenate(
                    [self.reference_window_y, self.cur_window_y], axis=0)
                self.reference_window_y = self.reference_window_y[max(
                    0, len(self.reference_window_y) - self.window_size):]

            self.cur_window_X = []
            self.cur_window_y = []
            self._fit_model()

        return self

    def _fit_model(self):
        self.reset_model()

        if self.reference_window_y is None:
            self.model.fit(self.reference_window_X)
        else:
            self.model.fit(self.reference_window_X, self.reference_window_y)

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            float: The anomalousness score of the input instance.
        """
        score = self.model.decision_function([X])[0]

        return score



# Copyright (c) 2017 Yusuke Sugomori
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# Portions of this code have been adapted from Yusuke Sugomori's code on GitHub: https://github.com/yusugomori/DeepLearning

import sys
import numpy
from .utils import *
import json

class dA_params:
    def __init__(self,n_visible = 5, n_hidden = 3, lr=0.001, corruption_level=0.0, gracePeriod = 10000, hiddenRatio=None):
        self.n_visible = n_visible# num of units in visible (input) layer
        self.n_hidden = n_hidden# num of units in hidden layer
        self.lr = lr
        self.corruption_level = corruption_level
        self.gracePeriod = gracePeriod
        self.hiddenRatio = hiddenRatio

class dA:
    def __init__(self, params):
        self.params = params

        if self.params.hiddenRatio is not None:
            self.params.n_hidden = int(numpy.ceil(self.params.n_visible*self.params.hiddenRatio))

        # for 0-1 normlaization
        self.norm_max = numpy.ones((self.params.n_visible,)) * -numpy.inf
        self.norm_min = numpy.ones((self.params.n_visible,)) * numpy.inf
        self.n = 0

        self.rng = numpy.random.RandomState(1234)

        a = 1. / self.params.n_visible
        self.W = numpy.array(self.rng.uniform(  # initialize W uniformly
            low=-a,
            high=a,
            size=(self.params.n_visible, self.params.n_hidden)))

        self.hbias = numpy.zeros(self.params.n_hidden)  # initialize h bias 0
        self.vbias = numpy.zeros(self.params.n_visible)  # initialize v bias 0
        self.W_prime = self.W.T


    def get_corrupted_input(self, input, corruption_level):
        assert corruption_level < 1

        return self.rng.binomial(size=input.shape,
                                 n=1,
                                 p=1 - corruption_level) * input

    # Encode
    def get_hidden_values(self, input):
        return sigmoid(numpy.dot(input, self.W) + self.hbias)

    # Decode
    def get_reconstructed_input(self, hidden):
        return sigmoid(numpy.dot(hidden, self.W_prime) + self.vbias)

    def train(self, x):
        self.n = self.n + 1
        # update norms
        self.norm_max[x > self.norm_max] = x[x > self.norm_max]
        self.norm_min[x < self.norm_min] = x[x < self.norm_min]

        # 0-1 normalize
        x = (x - self.norm_min) / (self.norm_max - self.norm_min + 0.0000000000000001)

        if self.params.corruption_level > 0.0:
            tilde_x = self.get_corrupted_input(x, self.params.corruption_level)
        else:
            tilde_x = x
        y = self.get_hidden_values(tilde_x)
        z = self.get_reconstructed_input(y)

        L_h2 = x - z
        L_h1 = numpy.dot(L_h2, self.W) * y * (1 - y)

        L_vbias = L_h2
        L_hbias = L_h1
        L_W = numpy.outer(tilde_x.T, L_h1) + numpy.outer(L_h2.T, y)

        self.W += self.params.lr * L_W
        self.hbias += self.params.lr * numpy.mean(L_hbias, axis=0)
        self.vbias += self.params.lr * numpy.mean(L_vbias, axis=0)
        return numpy.sqrt(numpy.mean(L_h2**2)) #the RMSE reconstruction error during training


    def reconstruct(self, x):
        y = self.get_hidden_values(x)
        z = self.get_reconstructed_input(y)
        return z

    def execute(self, x): #returns MSE of the reconstruction of x
        if self.n < self.params.gracePeriod:
            return 0.0
        else:
            # 0-1 normalize
            x = (x - self.norm_min) / (self.norm_max - self.norm_min + 0.0000000000000001)
            z = self.reconstruct(x)
            rmse = numpy.sqrt(((x - z) ** 2).mean()) #MSE
            return rmse


    def inGrace(self):
        return self.n < self.params.gracePeriod



from abc import ABC, abstractmethod
import numpy as np
from pysad.utils import _iterate


class BaseTransformer(ABC):
    """Base class for transforming methods.
    """

    def __init__(self, output_dims):
        self.output_dims = output_dims

    @abstractmethod
    def fit_partial(self, X):
        """Fits particular (next) timestep's features to train the transformer.

        Args:
            X (np.float64 array of shape (num_components,)): Input feature vector.
        Returns:
            object: self.
        """
        pass

    @abstractmethod
    def transform_partial(self, X):
        """Transforms particular (next) timestep's vector.

        Args:
            X (np.float64 array of shape (num_features,)): Input feature vector.

        Returns:
            transformed_X (np.float64 array of shape (num_components,)): Projected feature vector.
        """
        pass

    def fit_transform_partial(self, X):
        """Shortcut method that iteratively applies fit_partial and transform_partial, respectively.

        Args:
            X (np.float64 array of shape (num_components,)): Input feature vector.

        Returns:
            transformed_X (np.float64 array of shape (num_components,)): Projected feature vector.
        """
        return self.fit_partial(X).transform_partial(X)

    def transform(self, X):
        """Shortcut method that iteratively applies transform_partial to all instances in order.

        Args:
            X (np.float64 array of shape (num_instances, num_features)): Input feature vectors.

        Returns:
            np.float64 array of shape (num_instances, num_components): Projected feature vectors.
        """
        output_dims = self.output_dims if self.output_dims > 0 else X.shape[1]
        transformed_X = np.empty((X.shape[0], output_dims), dtype=np.float64)
        for i, (xi, _) in enumerate(_iterate(X)):
            result = self.transform_partial(xi)
            if np.asarray(result).ndim == 0:
                transformed_X[i] = result
            else:
                transformed_X[i] = result

        return transformed_X

    def fit(self, X):
        """Shortcut method that iteratively applies fit_partial to all instances in order.

        Args:
            X (np.float64 array of shape (num_instances, num_features)): Input feature vectors.

        Returns:
            object: The fitted transformer
        """
        for xi in _iterate(X):
            self.fit_partial(xi)

        return self

    def fit_transform(self, X):
        """Shortcut method that iteratively applies fit_transform_partial to all instances in order.

        Args:
            X (np.float64 array of shape (num_instances, num_components)): Input feature vectors.

        Returns:
            np.float64 array of shape (num_instances, num_components): Projected feature vectors.
        """
        output_dims = self.output_dims if self.output_dims > 0 else X.shape[1]
        transformed_X = np.empty((X.shape[0], output_dims), dtype=np.float64)
        for i, (xi, _) in enumerate(_iterate(X)):
            result = self.fit_transform_partial(xi)
            if np.asarray(result).ndim == 0:
                transformed_X[i] = result
            else:
                transformed_X[i] = result

        return transformed_X



"""
The :mod:`pysad.transform.projection` module contains methods to project input into (possibly) lower dimensional space to better discriminate anomalies.
"""
from .random_projector import BaseSKLearnProjector, GaussianRandomProjector, SparseRandomProjector
from .streamhash_projector import StreamhashProjector

__all__ = ["BaseSKLearnProjector", "GaussianRandomProjector", "SparseRandomProjector", "StreamhashProjector"]



from pysad.core.base_model import BaseModel


class PerfectModel(BaseModel):
    """This model directly outputs the ground truth labels. This method is added for testing and pipelining purposes.
    """

    def __init__(self):
        self.labels = []

    def fit_partial(self, X, y):
        """Fits the model to the ground truth label. Adds the label to the self.label queue.

        Args:
            X: any (Ignored)
            y (int): The true label.

        Returns:
            object: Returns the self.
        """
        if y is None:
            raise ValueError("y should be the true score")

        self.labels.append(y)

        return self

    def score_partial(self, X):
        """Pops the score from the self.label queue.

        Args:
            X: any (Ignored)

        Returns:
            float: The true label.
        """
        score = self.labels[0]
        self.labels = self.labels[1:]

        return score





def test_gaussian_random_projector(test_path):
    from pysad.transform.projection import GaussianRandomProjector

    for num_components in [2, 50, 250]:

        projector = GaussianRandomProjector(num_components=num_components)

        helper_test_projector(test_path, projector, num_components)


def test_sparse_random_projector(test_path):
    from pysad.transform.projection import SparseRandomProjector

    for num_components in [2, 50, 250]:

        projector = SparseRandomProjector(num_components=num_components)

        helper_test_projector(test_path, projector, num_components)


def helper_test_projector(test_path, projector, num_components):
    import os
    from sklearn.utils import shuffle
    from pysad.utils import Data

    data = Data(os.path.join(test_path, "../../../examples/data"))

    X_all, y_all = data.get_data("arrhythmia.mat")
    X_all, y_all = shuffle(X_all, y_all)
    projected_X = projector.fit_transform(X_all)

    assert projected_X.shape == (X_all.shape[0], num_components)



from pysad.statistics.abs_statistic import AbsStatistic
from pysad.statistics.running_statistic import RunningStatistic


def test_all_zero_stats():
    import numpy as np
    from pysad.statistics import AbsStatistic
    from pysad.statistics import RunningStatistic
    from pysad.statistics import AverageMeter
    from pysad.statistics import CountMeter
    from pysad.statistics import MaxMeter
    from pysad.statistics import MedianMeter
    from pysad.statistics import MinMeter
    from pysad.statistics import SumMeter
    from pysad.statistics import SumSquaresMeter
    from pysad.statistics import VarianceMeter
    from pysad.utils import fix_seed
    fix_seed(61)

    num_items = 100
    stat_classes = {
        AverageMeter: 0.0,
        CountMeter: "count",
        MaxMeter: 0.0,
        MedianMeter: 0.0,
        MinMeter: 0.0,
        SumMeter: 0.0,
        SumSquaresMeter: 0.0,
        VarianceMeter: 0.0
    }

    for stat_cls, val in stat_classes.items():
        stat = stat_cls()
        abs_stat = AbsStatistic(stat_cls)
        window_size = 25
        running_stat = RunningStatistic(stat_cls, window_size=window_size)
        arr = np.zeros(num_items, dtype=np.float64)
        prev_value = 0.0
        for i in range(arr.shape[0]):
            num = arr[i]
            stat.update(num)
            abs_stat.update(num)
            running_stat.update(num)
            if i > 1: # for variance meter.
                assert np.isclose(stat.get(), val if val != "count" else i+1)
                assert np.isclose(abs_stat.get(), val if val != "count" else i+1)
                assert np.isclose(running_stat.get(), val if val != "count" else min(i+1, window_size))

                stat.remove(num)
                abs_stat.remove(num)
                assert np.isclose(stat.get(), prev_value)
                assert np.isclose(abs_stat.get(), abs(prev_value))
                stat.update(num)
                abs_stat.update(num)

            prev_value = stat.get()


def test_stats_with_batch_numpy():

    from pysad.statistics import AverageMeter
    from pysad.statistics import CountMeter
    from pysad.statistics import MaxMeter
    from pysad.statistics import MedianMeter
    from pysad.statistics import MinMeter
    from pysad.statistics import SumMeter
    from pysad.statistics import SumSquaresMeter
    from pysad.statistics import VarianceMeter
    import numpy as np
    from pysad.utils import fix_seed
    fix_seed(61)

    num_items = 100
    stat_classes = {
        AverageMeter: np.mean,
        CountMeter: len,
        MaxMeter: np.max,
        MedianMeter: np.median,
        MinMeter: np.min,
        SumMeter: np.sum,
        SumSquaresMeter: lambda x: np.sum(x**2),
        VarianceMeter: np.var
    }

    for stat_cls, val in stat_classes.items():
        stat = stat_cls()
        abs_stat = AbsStatistic(stat_cls)
        window_size = 25
        running_stat = RunningStatistic(stat_cls, window_size=window_size)

        arr = np.random.rand(num_items)
        prev_value = 0.0
        for i in range(arr.shape[0]):
            num = arr[i]
            stat.update(num)
            abs_stat.update(num)
            running_stat.update(num)

            if i > 1: # for variance meter.
                assert np.isclose(stat.get(), val(arr[:i+1]))
                assert np.isclose(running_stat.get(), val(arr[max(0, i-window_size+1):i+1]))
                assert np.isclose(abs(stat.get()), abs_stat.get())

            stat.remove(num)
            abs_stat.remove(num)

            if i > 1:
                assert np.isclose(stat.get(), prev_value)
                assert np.isclose(abs_stat.get(), abs(prev_value))

            stat.update(num)
            abs_stat.update(num)

            prev_value = stat.get()



from abc import abstractmethod
from pyod.models.combination import average, maximization, median, moa, aom
from pysad.core.base_postprocessor import BasePostprocessor


class PYODScoreEnsembler(BasePostprocessor):
    """Abstract base class for the scoring ensembling methods for the scoring based ensemblers of the `PyOD <https://pyod.readthedocs.io/en/latest/pyod.models.html#module-pyod.models.combination>`_.
    """

    @abstractmethod
    def _combine(self, scores):
        """Abstract method that directly uses  of our framework to be filled.

        Args:
            scores: np.float64 array of shape (1, num_scores)
                List of scores from multiple anomaly detectors.

        Returns:
            float: Resulting anomaly score.
        """
        pass

    def fit_partial(self, scores):
        """Fits particular (next) timestep's score to train the ensembler. For PYOD based ensemblers, this method does not affect anything and returns self directly.

        Args:
            scores: np.float64 array of shape (num_anomaly_detectors, )
                List of scores from multiple anomaly detectors.

        Returns:
            object: The fitted ensembler.
        """
        return self

    def transform_partial(self, scores):
        """Combines anomaly scores from multiple anomaly detectors for a particular timestep.

        Args:
            scores: np.float64 array of shape (num_anomaly_detectors, )
                List of scores from multiple anomaly detectors.

        Returns:
            float: Resulting anomaly score.
        """
        scores = scores.reshape(1, -1)

        return self._combine(scores)


class AverageScoreEnsembler(PYODScoreEnsembler):
    """An wrapper class that results in the weighted average of the anomaly scores from multiple anomaly detectors. For more details, see `PyOD documentation <https://pyod.readthedocs.io/en/latest/pyod.models.html#module-pyod.models.combination>`_.

        Args:
            estimator_weights (np array of shape (1, num_anomaly_detectors)): The weights for detectors. If None, uniform weights are assigned.

    """

    def __init__(self, estimator_weights=None):

        self.estimator_weights = estimator_weights

    def _combine(self, scores):
        """Wrapping for PyOD the ensembler.

        Args:
            scores (np.float64 array of shape (num_anomaly_detectors, )): List of scores from multiple anomaly detectors.

        Returns:
            float: Resulting anomaly score.
        """
        return average(scores, estimator_weights=self.estimator_weights)


class MaximumScoreEnsembler(PYODScoreEnsembler):
    """An ensembler that results the maximum of the previous scores.
    """

    def _combine(self, scores):
        """
        Wrapping for PyOD the ensembler.
        Args:
            scores (np.float64 array of shape (num_anomaly_detectors, )) List of scores from multiple anomaly detectors.

        Returns:
            float: Resulting anomaly score.
        """
        return maximization(scores)


class MedianScoreEnsembler(PYODScoreEnsembler):
    """An ensembler that results the median of the previous scores.
    """

    def _combine(self, scores):
        """
        Helper method to wrap the PyOD ensembler.
        Args:
            scores (np.float64 array of shape (num_anomaly_detectors, )) : List of scores from multiple anomaly detectors.

        Returns:
            float: Resulting anomaly score.
        """
        return median(scores)


class AverageOfMaximumScoreEnsembler(PYODScoreEnsembler):
    """Maximum of average scores ensembler that outputs the maximum of average. For more details, see :cite:`aggarwal2015theoretical` and `PyOD documentation <https://pyod.readthedocs.io/en/latest/pyod.models.html#module-pyod.models.combination>`_. The ensembler firt divides the scores into buckets and takes the maximum for each bucket. Then, the ensembler outputs the average of all these maximum scores of buckets.

    Args:
        n_buckets (int): The number of subgroups to build (Default=5).
        method (str):  {'static', 'dynamic'}, if 'dynamic', build subgroups randomly with dynamic bucket size (Default='static').
        bootstrap_estimators (bool) Whether estimators are drawn with replacement (Default=False).
    """

    def __init__(
            self,
            n_buckets=5,
            method='static',
            bootstrap_estimators=False):
        self.method = method
        self.n_buckets = n_buckets
        self.bootstrap_estimators = bootstrap_estimators

    def _combine(self, scores):
        """Wrapping for PyOD the ensembler.

        Args:
            scores (np.float64 array of shape (num_anomaly_detectors, )): List of scores from multiple anomaly detectors.

        Returns:
            float: Resulting anomaly score.
        """
        return aom(
            scores,
            n_buckets=self.n_buckets,
            method=self.method,
            bootstrap_estimators=self.bootstrap_estimators)


class MaximumOfAverageScoreEnsembler(PYODScoreEnsembler):
    """Maximum of average scores ensembler that outputs the maximum of average. For more details, see :cite:`aggarwal2015theoretical` and `PyOD documentation <https://pyod.readthedocs.io/en/latest/pyod.models.html#module-pyod.models.combination>`_. The ensembler firt divides the scores into buckets and takes the average for each bucket. Then, the ensembler outputs the maximum of all these average scores of buckets.

    Args:
        n_buckets  : int, optional (default=5)
            The number of subgroups to build
        method (str): {'static', 'dynamic'}, if 'dynamic', build subgroups randomly with dynamic bucket size (default='static').
        bootstrap_estimators (bool): Whether estimators are drawn with replacement (Default=False).
    """

    def __init__(
            self,
            n_buckets=5,
            method='static',
            bootstrap_estimators=False):
        self.method = method
        self.n_buckets = n_buckets
        self.bootstrap_estimators = bootstrap_estimators

    def _combine(self, scores):
        """
        Wrapping for PyOD the ensembler.
        Args:
            scores: np.float64 array of shape (num_anomaly_detectors, )
                List of scores from multiple anomaly detectors.

        Returns:
            float: Resulting anomaly score.
        """
        return moa(
            scores,
            n_buckets=self.n_buckets,
            method=self.method,
            bootstrap_estimators=self.bootstrap_estimators)



"""
The :mod:`pysad.statistics` module contains methods to keep track of statistics on streaming data.
"""
from .abs_statistic import AbsStatistic
from .average_meter import AverageMeter
from .count_meter import CountMeter
from .max_meter import MaxMeter
from .median_meter import MedianMeter
from .min_meter import MinMeter
from .running_statistic import RunningStatistic
from .sum_meter import SumMeter
from .sum_squares_meter import SumSquaresMeter
from .variance_meter import VarianceMeter

__all__ = ["AbsStatistic", "AverageMeter", "CountMeter", "MaxMeter", "MedianMeter", "MinMeter", "RunningStatistic", "SumMeter", "SumSquaresMeter", "VarianceMeter"]



import numpy as np
import os
import tempfile
from pysad.utils.data import Data


def test_get_data_mat(monkeypatch):
    # Simulate scipy.io.loadmat
    class DummyF:
        def __getitem__(self, key):
            if key == 'X':
                return np.ones((5, 2))
            if key == 'y':
                return np.arange(5).reshape(-1, 1)
    def dummy_loadmat(path):
        return DummyF()
    monkeypatch.setattr("scipy.io.loadmat", dummy_loadmat)
    d = Data(data_base_path=".")
    X, y = d.get_data("dummy.mat")
    assert X.shape == (5, 2)
    assert np.all(y == np.arange(5))


def test_get_data_txt():
    # Create a temporary .txt file
    arr = np.hstack([np.random.rand(5, 2), np.arange(5).reshape(-1, 1)])
    with tempfile.NamedTemporaryFile(mode="w+t", suffix=".txt", delete=False) as f:
        np.savetxt(f, arr, delimiter=",")
        fname = os.path.basename(f.name)
        f.close()
        d = Data(data_base_path=os.path.dirname(f.name))
        X, y = d.get_data(fname)
        assert X.shape == (5, 2)
        assert np.allclose(y, np.arange(5))
    os.remove(f.name)


def test_get_iterator(monkeypatch):
    # Patch get_data to return fixed arrays
    d = Data()
    d.get_data = lambda data_file: (np.ones((3, 2)), np.arange(3))
    it = d.get_iterator("dummy.txt", shuffle=False)
    items = list(it)
    assert len(items) == 3
    for x, y in items:
        assert np.all(x == 1)
        assert y in [0, 1, 2]


def test_get_data_files():
    d = Data()
    files = d._get_data_files()
    assert isinstance(files, list)
    assert "arrhythmia.mat" in files


def test_get_iterator_with_seed(monkeypatch):
    """Test get_iterator with seed parameter to ensure np.random.seed is called."""
    d = Data()
    d.get_data = lambda data_file: (np.ones((3, 2)), np.arange(3))
    
    # Mock np.random.seed to track if it's called
    seed_called = []
    original_seed = np.random.seed
    def mock_seed(seed):
        seed_called.append(seed)
        original_seed(seed)
    
    monkeypatch.setattr("numpy.random.seed", mock_seed)
    
    # Test with seed
    it = d.get_iterator("dummy.txt", shuffle=False, seed=42)
    list(it)  # Consume iterator
    assert 42 in seed_called


def test_get_iterator_without_seed(monkeypatch):
    """Test get_iterator without seed parameter."""
    d = Data()
    d.get_data = lambda data_file: (np.ones((3, 2)), np.arange(3))
    
    # Should work without seed
    it = d.get_iterator("dummy.txt", shuffle=False)
    items = list(it)
    assert len(items) == 3


def test_data_init_default_path():
    """Test Data class initialization with default path."""
    d = Data()
    assert d.data_base_path == "data"


def test_data_init_custom_path():
    """Test Data class initialization with custom path."""
    d = Data(data_base_path="/custom/path")
    assert d.data_base_path == "/custom/path"


def test_load_via_txt_method():
    """Test the _load_via_txt method directly."""
    # Create a temporary .txt file
    arr = np.random.rand(4, 3)
    with tempfile.NamedTemporaryFile(mode="w+t", suffix=".txt", delete=False) as f:
        np.savetxt(f, arr, delimiter=",")
        f.close()
        
        d = Data()
        result = d._load_via_txt(f.name)
        assert result.shape == (4, 3)
        assert np.allclose(result, arr)
    
    os.remove(f.name)


def test_get_data_files_content():
    """Test that _get_data_files returns expected file names."""
    d = Data()
    files = d._get_data_files()
    
    # Check some specific files are present
    expected_files = [
        'arrhythmia.mat', 'cardio.mat', 'glass.mat', 
        'gisette_sampled.txt', 'pima-indians_sampled.txt'
    ]
    
    for expected_file in expected_files:
        assert expected_file in files
    
    # Check that all files have proper extensions
    for file in files:
        assert file.endswith('.mat') or file.endswith('.txt')



from pysad.models.integrations.reference_window_model import ReferenceWindowModel
from pyod.models.iforest import IForest


class IForestASD(ReferenceWindowModel):
    """An Anomaly Detection Approach Based on Isolation Forest Algorithm for Streaming Data using Sliding Window :cite:`ding2013anomaly`. Note that concept drift is not implemented since it is a part of the simulation. See Algorithm 2 in "An Anomaly Detection Approach Based on Isolation Forest Algorithm for Streaming Data using Sliding Window" paper. This method is unsupervised so it is not needed to give y as parameter.

    Args:
        initial_window_X (np.float64 array of shape (num_initial_instances,num_features)): The initial window to fit for initial calibration period. We simply apply fit to these instances (Default=None).
        window_size (int): The size of the reference window and its sliding (Default=2048).
    """
    def __init__(self, initial_window_X=None, window_size=2048, **kwargs):
        super().__init__(IForest, window_size, window_size, initial_window_X, **kwargs)
        # TODO: implement concept drift method



from pysad.transform.postprocessing import AveragePostprocessor, MaxPostprocessor, MedianPostprocessor, \
    ZScorePostprocessor
from pysad.transform.postprocessing import RunningZScorePostprocessor, \
    RunningMedianPostprocessor, RunningMaxPostprocessor, RunningAveragePostprocessor
import warnings


def helper_get_scores():
    import numpy as np
    # Use a fixed seed for reproducible results and avoid edge cases
    np.random.seed(42)
    # Create more diverse scores to avoid variance issues
    scores = np.random.rand(100) * 10 + np.arange(100) * 0.1

    return scores


def test_postprocessors_shape():
    scores = helper_get_scores()

    postprocessors = {
        AveragePostprocessor: {},
        MaxPostprocessor: {},
        MedianPostprocessor: {},
        ZScorePostprocessor: {},
        RunningAveragePostprocessor: { "window_size": 30 },
        RunningMaxPostprocessor: { "window_size": 30 },
        RunningMedianPostprocessor: { "window_size": 30 },
        RunningZScorePostprocessor: { "window_size": 30 },
    }

    for postprocessor_cls, params_dict in postprocessors.items():
        postprocessor = postprocessor_cls(**params_dict)
        # Suppress RuntimeWarning for division by zero in z-score calculations
        # This can happen with small variance values or edge cases
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning, 
                                  message="invalid value encountered in scalar divide")
            postprocessed_scores = postprocessor.fit_transform(scores)
        assert scores.shape == postprocessed_scores.shape



"""
An open-source python framework for anomaly detection on streaming multivariate data.
"""
import warnings

# Filter out NumPy deprecation warnings to avoid noise in user applications
warnings.filterwarnings("ignore", message="numpy.core is deprecated")
warnings.filterwarnings("ignore", message="Conversion of an array with ndim > 0 to a scalar is deprecated")
warnings.filterwarnings("ignore", message="Calling nonzero on 0d arrays is deprecated")

from .version import __version__
from . import core, evaluation, models, statistics, transform, utils

__all__ = ['__version__', 'core', "evaluation", "models", "statistics", "transform", "utils"]



import copy
import numpy as np
from pysad.core.base_model import BaseModel


class HalfSpaceTrees(BaseModel):
    """Half-Space Trees method :cite:`tan2011fast`.

    Args:
        feature_mins (np.float64 array of shape (num_features,)): Minimum boundary of the features.
        feature_maxes (np.float64 array of shape (num_features,)): Maximum boundary of the features.
        window_size (int): The size of the window (Default=100).
        num_trees (int): The number of treesint (Default=25).
        max_depth (int): Maximum depth of the trees (Default=15).
        initial_window_X (np.float64 array of shape (num_initial_instances,num_features)): The initial window to fit for initial calibration period. If not `None`, we simply apply fit to these instances (Default=None).
    """

    def __init__(
            self,
            feature_mins,
            feature_maxes,
            window_size=100,
            num_trees=25,
            max_depth=15,
            initial_window_X=None):
        self.window_size = window_size
        self.max_depth = max_depth
        self.num_trees = num_trees
        self.feature_maxes = feature_maxes
        self.feature_mins = feature_mins

        self.num_dimensions = len(self.feature_maxes)

        self.roots = [
            self._build_single_hs_tree(
                copy.deepcopy(
                    self.feature_mins),
                copy.deepcopy(
                    self.feature_maxes),
                0) for _ in range(
                self.num_trees)]

        self.is_first_window = True
        self.step = 0
        if initial_window_X:
            self.fit(initial_window_X)

    def _build_single_hs_tree(self, mins, maxes, current_depth):
        if current_depth == self.max_depth:
            return self._Node(
                left=None,
                right=None,
                r_mass=0,
                l_mass=0,
                split_att=0,
                split_value=0.0,
                k=current_depth)

        q = np.random.randint(self.num_dimensions)
        p = (maxes[q] + mins[q]) / 2.0

        temp = maxes[q]
        maxes[q] = p
        left = self._build_single_hs_tree(
            copy.deepcopy(mins),
            copy.deepcopy(maxes),
            current_depth + 1)
        maxes[q] = temp
        mins[q] = p
        right = self._build_single_hs_tree(
            copy.deepcopy(mins),
            copy.deepcopy(maxes),
            current_depth + 1)

        return self._Node(
            left=left,
            right=right,
            r_mass=0,
            l_mass=0,
            split_att=q,
            split_value=p,
            k=current_depth)

    def _update_mass(self, x, node, ref_window):
        if ref_window:
            node.r_mass += 1
            node.l_mass += 1  # Does not exist in original since we want it to predict while building the first window
        else:
            node.l_mass += 1

        if node.k < self.max_depth:
            target_node = node.right if x[node.split_att] > node.split_value else node.left

            self._update_mass(x, target_node, ref_window)

    def _update_model(self, node):

        if node is None:
            return

        self.is_first_window = False
        node.r_mass = node.l_mass
        node.l_mass = 0

        self._update_model(node.left)
        self._update_model(node.right)

    def fit_partial(self, X, y=None):
        """Fits the model to next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to fit.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: Returns the self.
        """
        self.step += 1

        for root in self.roots:
            self._update_mass(X, root, self.is_first_window)

        if self.step % self.window_size == 0:
            for root in self.roots:
                self._update_model(root)

        return self

    def _score_tree(self, X, node):
        if node is None:
            return 0.0

        target_node = node.right if X[node.split_att] > node.split_value else node.left

        return node.r_mass * (2**node.k) + self._score_tree(X, target_node)

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            float: The anomalousness score of the input instance.
        """
        s = 0.0

        for root in self.roots:
            s += self._score_tree(X, root)

        return -s

    class _Node:
        def __init__(self, left, right, r_mass, l_mass, split_att, split_value, k):
            self.left = left
            self.right = right
            self.r_mass = r_mass
            self.l_mass = l_mass
            self.split_att = split_att
            self.split_value = split_value
            self.k = k



from pysad.core.base_model import BaseModel


class PYODModel(BaseModel):

    def __init__(self, model_cls, **kwargs):
        """Abstract base class for PYOD models.

        Args:
            model_cls (class): The model class to be instantiated.
            **kwargs (Keyword arguments): Keyword arguments that is passed to the `model_cls`.
        """
        self.model_cls = model_cls
        self.kwargs = kwargs
        self.model = None

    def reset_model(self):
        """Removes the old model from the memory and instantiates a new one.
        """
        self.model = self.model_cls(**self.kwargs)



from pysad.core.base_statistic import BaseStatistic


class RunningStatistic(BaseStatistic):
    """The running statistic that wraps any other statistics to track statistics with a fixed window size.

    Args:
        statistic_cls (class): The class to be instantiated and to be windowed.
        window_size (int): The window size.
        **kwargs (Keyword arguments): The keyword arguments that is input to the statistic_cls.
    """

    def __init__(self, statistic_cls, window_size, **kwargs):
        super().__init__(**kwargs)
        self.statistic_cls = statistic_cls
        self.statistic = self.statistic_cls()

        self.window_size = window_size
        self.window = []

    def update(self, num):
        """Updates the statistic with the value for a timestep.

        Args:
            num (float): The incoming value, for which the statistic is used.

        Returns:
            object: self.
        """
        self.window.append(num)

        self.statistic.update(num)

        if len(self.window) > self.window_size:
            self.statistic.remove(self.window[0])
            self.window = self.window[1:]

        return self

    def get(self):
        """ Method to obtain the tracked statistic.

        Returns:
            float: The statistic.
        """
        return self.statistic.get()




import numpy
from scipy.stats import norm

def pdf(x,mu,sigma): #normal distribution pdf
    x = (x-mu)/sigma
    return numpy.exp(-x**2/2)/(numpy.sqrt(2*numpy.pi)*sigma)

def invLogCDF(x,mu,sigma): #normal distribution cdf
    x = (x - mu) / sigma
    return norm.logcdf(-x) #note: we mutiple by -1 after normalization to better get the 1-cdf

def sigmoid(x):
    return 1. / (1 + numpy.exp(-x))


def dsigmoid(x):
    return x * (1. - x)

def tanh(x):
    return numpy.tanh(x)

def dtanh(x):
    return 1. - x * x

def softmax(x):
    e = numpy.exp(x - numpy.max(x))  # prevent overflow
    if e.ndim == 1:
        return e / numpy.sum(e, axis=0)
    else:  
        return e / numpy.array([numpy.sum(e, axis=1)]).T  # ndim = 2


def ReLU(x):
    return x * (x > 0)

def dReLU(x):
    return 1. * (x > 0)

class rollmean:
    def __init__(self,k):
        self.winsize = k
        self.window = numpy.zeros(self.winsize)
        self.pointer = 0

    def apply(self,newval):
        self.window[self.pointer]=newval
        self.pointer = (self.pointer+1) % self.winsize
        return numpy.mean(self.window)

# probability density for the Gaussian dist
# def gaussian(x, mean=0.0, scale=1.0):
#     s = 2 * numpy.power(scale, 2)
#     e = numpy.exp( - numpy.power((x - mean), 2) / s )

#     return e / numpy.square(numpy.pi * s)




from pysad.core.base_model import BaseModel


class LocalOutlierProbability(BaseModel):
    """The implementation of streaming Local Outlier Probabilities method :cite:`kriegel2009loop`, which uses the implementation of PyNomaly library :cite:`constantinou2018pynomaly`.

    Args:
        initial_X (np.float64 array of shape (num_instances, num_features)): Initial training data to calibrate the model.
        num_neighbors (int): Number of neighbors (Default=10).
        extent (int): an integer value that controls the statistical extent, e.g. lambda times the standard deviation from the mean (optional, default 3)
        n_neighbors (int): the total number of neighbors to consider w.r.t. each sample (optional, default 10)
    """

    def __init__(self, initial_X, num_neighbors=10, extent=3):
        from PyNomaly import loop

        self.model = loop.LocalOutlierProbability(
            initial_X, extent=extent, n_neighbors=num_neighbors).fit()

    def fit_partial(self, X, y=None):
        """Fits the model to next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to fit.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: self.
        """
        return self

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            float: The anomalousness score of the input instance.
        """
        return self.model.stream(X)



from pysad.core.base_transformer import BaseTransformer


class InstanceStandardScaler(BaseTransformer):
    """Standard deviation scaling per instance. Not that the variance and mean is calculated per instance, for which the scaling is done with.
    The method substracts mean and divides with the standard deviation of the features, separately for each instance.
    """

    def __init__(self):
        super().__init__(-1)

    def fit_partial(self, X):
        """Fits particular (next) timestep's features to train the scaler.

        Args:
            X (np.float64 array of shape (num_features,)): Input feature vector.
        Returns:
            object: self.
        """
        return self

    def transform_partial(self, X):
        """Scales particular (next) timestep's vector.

        Args:
            X (np.float64 array of shape (num_features,)): Input feature vector.

        Returns:
            scaled_X (np.float64 array of shape (features,)): Scaled feature vector.
        """

        X_mean = X.mean()
        X_std = X.std()

        return (X - X_mean) / X_std



import pandas as pd
import numpy as np
import pytest
from pysad.utils.pandas_streamer import PandasStreamer


def test_pandas_streamer_iter_no_labels():
    df = pd.DataFrame(np.random.rand(10, 3))
    streamer = PandasStreamer(shuffle=False)
    result = list(streamer.iter(df))
    assert len(result) == 10
    assert all(isinstance(row, np.ndarray) for row in result)
    assert all(row.shape == (3,) for row in result)


def test_pandas_streamer_iter_with_labels():
    df = pd.DataFrame(np.random.rand(10, 3))
    labels = pd.Series(np.arange(10))
    streamer = PandasStreamer(shuffle=False)
    result = list(streamer.iter(df, labels))
    assert len(result) == 10
    for x, y in result:
        assert isinstance(x, np.ndarray)
        assert x.shape == (3,)
        assert isinstance(y, (int, np.integer, float, np.floating))


def test_pandas_streamer_iter_length_mismatch():
    df = pd.DataFrame(np.random.rand(10, 3))
    labels = pd.Series(np.arange(9))
    streamer = PandasStreamer(shuffle=False)
    with pytest.raises(AssertionError):
        list(streamer.iter(df, labels))


def test_pandas_streamer_shuffle():
    """Test that shuffle parameter works correctly."""
    df = pd.DataFrame(np.arange(20).reshape(10, 2))

    # Test with shuffle=True
    streamer_shuffle = PandasStreamer(shuffle=True)
    result_shuffle = list(streamer_shuffle.iter(df))

    # Test with shuffle=False
    streamer_no_shuffle = PandasStreamer(shuffle=False)
    result_no_shuffle = list(streamer_no_shuffle.iter(df))

    # Both should have same length
    assert len(result_shuffle) == len(result_no_shuffle) == 10

    # With shuffle=False, first row should be [0, 1]
    assert np.array_equal(result_no_shuffle[0], np.array([0, 1]))


def test_pandas_streamer_inheritance():
    """Test that PandasStreamer properly inherits from BaseStreamer."""
    streamer = PandasStreamer(shuffle=True)
    assert streamer.shuffle is True

    streamer = PandasStreamer(shuffle=False)
    assert streamer.shuffle is False


def test_pandas_streamer_empty_dataframe():
    """Test with empty dataframe."""
    df = pd.DataFrame()
    streamer = PandasStreamer(shuffle=False)
    result = list(streamer.iter(df))
    assert len(result) == 0





def test_instance_standard_scaler():
    import numpy as np
    from pysad.transform.preprocessing import InstanceStandardScaler

    X = np.random.rand(100, 25)
    scaler = InstanceStandardScaler()

    scaled_X = scaler.fit_transform(X)
    assert np.all(np.isclose(scaled_X, (X - X.mean(1, keepdims=True))/X.std(1, keepdims=True)))

    scaler = scaler.fit(X)
    scaled_X = scaler.transform(X)
    assert np.all(np.isclose(scaled_X, (X - X.mean(1, keepdims=True))/X.std(1, keepdims=True)))



"""
The :mod:`pysad.core` module covers base classes of the `PySAD`.
"""
from .base_metric import BaseMetric
from .base_model import BaseModel
from .base_postprocessor import BasePostprocessor
from .base_statistic import BaseStatistic
from .base_streamer import BaseStreamer
from .base_transformer import BaseTransformer

__all__ = ["BaseMetric", "BaseModel", "BasePostprocessor", "BaseStatistic", "BaseStreamer", "BaseTransformer"]



from pysad.core.base_statistic import UnivariateStatistic


class SumSquaresMeter(UnivariateStatistic):
    """The statistic that keeps track of the sum of squares. This class is useful to calculate and used in VarianceMeter.

    Attrs:
        sum_squares (float): The sum of squares of values.
        num_items (int): The number of items that are used to update the statistic.
    """

    def __init__(self):
        self.sum_squares = 0.0
        self.num_items = 0

    def update(self, num):
        """Updates the statistic with the value for a timestep.

        Args:
            num (float): The incoming value, for which the statistic is used.

        Returns:
            object: self.
        """
        self.sum_squares += num**2
        self.num_items += 1

        return self

    def remove(self, num):
        """Updates the statistic by removing particular value. This method

        Args:
            num (float): The value to be removed.

        Returns:
            object: self.
        """
        self.sum_squares -= num**2
        self.num_items -= 1

        return self

    def get(self):
        """ Method to obtain the tracked statistic.

        Returns:
            float: The statistic.
        """
        return self.sum_squares



from pysad.core.base_model import BaseModel
import numpy as np


class KNNCAD(BaseModel):
    """Conformalized density- and distance-based anomaly detection in time-series data :cite:`burnaev2016conformalized`, which uses a combination of a feature extraction method, an approach to assess a score whether a new observation differs significantly from a previously observed data, and a probabilistic interpretation of this score based on the conformal paradigm. This method's implementation is based on `NAB-kNNCAD <https://github.com/numenta/NAB/blob/master/nab/detectors/knncad/knncad_detector.py>`_. This model is univariate.

        Args:
            probationary_period (int): Number of instances in probationary period. Until probationary_period instances are received, the model outputs anomaly score of `0.0`.
    """

    def __init__(self, probationary_period):
        self.buf = []
        self.training = []
        self.calibration = []
        self.scores = []
        self.record_count = 0
        self.pred = -1
        self.k = 27
        self.to_init = True

        self.probationaryPeriod = probationary_period

    def _metric(self, a, b):
        diff = a - np.array(b)

        return np.dot(np.dot(diff, self.sigma), diff.T)

    def _ncm(self, item, item_in_array=False):
        arr = [self._metric(x, item) for x in self.training]

        return np.sum(np.partition(arr, self.k + item_in_array)[:self.k + item_in_array])

    def fit_partial(self, X, y=None):
        """Fits the model to next instance. Note that this model is univariate.

        Args:
            X (np.float64 array of shape (1,)): The instance to fit.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: Returns the self.
        """
        if self.to_init:
            self.dim = 19  # X.shape[0]
            self.sigma = np.diag(np.ones(self.dim))
            self.to_init = False

        self.buf.append(X[0])
        self.record_count += 1
        if len(self.buf) < self.dim:
            return self

        new_item = self.buf[-self.dim:]

        if self.record_count < self.probationaryPeriod:
            self.training.append(new_item)
        else:
            ost = self.record_count % self.probationaryPeriod
            if ost == 0 or ost == int(self.probationaryPeriod / 2):
                try:
                    self.sigma = np.linalg.inv(
                        np.dot(np.array(self.training).T, self.training))
                except np.linalg.linalg.LinAlgError:
                    print('Singular Matrix at record', self.record_count)
            if len(self.scores) == 0:
                self.scores = [self._ncm(v, True) for v in self.training]

            new_score = self._ncm(new_item)

            if self.record_count >= 2 * self.probationaryPeriod:
                self.training.pop(0)
                self.training.append(self.calibration.pop(0))

            self.scores.pop(0)
            self.calibration.append(new_item)
            self.scores.append(new_score)

        return self

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (1,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            float: The anomalousness score of the input instance.
        """
        if len(self.buf) < self.dim or self.record_count < self.probationaryPeriod:
            return 0.0

        new_item = self.buf[-self.dim:]

        new_score = self._ncm(new_item)
        result = 1. * len(np.where(np.array(self.scores) < new_score)[0]) / len(self.scores)

        if self.pred > 0:
            self.pred -= 1
            return 0.5
        elif result >= 0.9965:
            self.pred = int(self.probationaryPeriod / 5)

        return result



import abc
from abc import abstractmethod


class BaseStreamer(abc.ABC):
    """Abstract base class to simulate the streaming data.

    Args:
        shuffle (bool): Whether shuffle the data initially (Optional, default=False).
    """

    def __init__(self, shuffle=False):
        self.shuffle = shuffle

    @abstractmethod
    def iter(self, X, y=None):
        """Method that iterates array of data and (optionally) labels.

        Args:
            X (np.array of shape (num_instances, num_features)): The features of instances to iterate.
            y: (Optional, default=None) If not None, iterates labels with the same order.
        """
        pass





def test_streamhash_projector(test_path):
    from sklearn.utils import shuffle
    from pysad.utils import Data
    import os
    from pysad.transform.projection import StreamhashProjector

    for num_components in [2, 50, 250]:
        data = Data(os.path.join(test_path, "../../../examples/data"))

        X_all, y_all = data.get_data("arrhythmia.mat")
        X_all, y_all = shuffle(X_all, y_all)

        projector = StreamhashProjector(num_components=num_components)

        projected_X = projector.fit_transform(X_all)

        assert projected_X.shape == (X_all.shape[0], num_components)



def helper_test_all_metrics(metric_classes, y_true, y_pred):
    import numpy as np

    for metric_cls, val in metric_classes.items():
        metric = metric_cls()

        for i, (yt, yp) in enumerate(zip(y_true, y_pred)):
            metric.update(yt, yp)
            if i > 0:
                assert np.isclose(metric.get(), val)


def test_all_correct():
    from pysad.evaluation import PrecisionMetric, AUPRMetric, AUROCMetric, RecallMetric
    import numpy as np
    from pysad.utils import fix_seed
    fix_seed(61)

    metric_classes = [
        PrecisionMetric,
        RecallMetric,
        AUPRMetric,
        AUROCMetric
    ]
    metric_classes = { metric_cls: 1.0 for metric_cls in metric_classes }
    y_true = np.random.randint(0, 2, size=(25,), dtype=np.int32)
    y_true[0] = 1
    y_true[1] = 0
    y_pred = y_true.copy()

    helper_test_all_metrics(metric_classes, y_true, y_pred)


def test_none_correct():
    from pysad.evaluation import PrecisionMetric, AUPRMetric, AUROCMetric, RecallMetric
    import numpy as np
    from pysad.utils import fix_seed
    fix_seed(61)

    metric_classes = {
        PrecisionMetric: 0.0,
        #AUPRMetric: 0.5
        AUROCMetric: 0.0,
        RecallMetric: 0.0
    }
    y_true = np.random.randint(0, 2, size=(25,), dtype=np.int32)
    y_true[0] = 1
    y_true[1] = 0
    y_pred = 1 - y_true.copy()

    helper_test_all_metrics(metric_classes, y_true, y_pred)


def test_base_sklearn_metric():
    """Test BaseSKLearnMetric functionality."""
    from pysad.evaluation import BaseSKLearnMetric
    from sklearn.metrics import accuracy_score
    
    class TestSKLearnMetric(BaseSKLearnMetric):
        def _evaluate(self, y_true, y_pred):
            return accuracy_score(y_true, y_pred)
    
    metric = TestSKLearnMetric()
    
    # Test with data
    metric.update(1, 1)  # Correct
    metric.update(0, 0)  # Correct  
    metric.update(1, 0)  # Incorrect
    
    # Should have 2/3 accuracy
    accuracy = metric.get()
    assert abs(accuracy - 2.0/3.0) < 1e-10


def test_metric_error_handling():
    """Test metric error handling with invalid data."""
    from pysad.evaluation import AUROCMetric
    
    metric = AUROCMetric()
    
    # Test with both classes present
    metric.update(1, 0.5)
    metric.update(0, 0.7)
    metric.update(1, 0.3)
    
    # AUROC should work with both classes
    score = metric.get()
    assert isinstance(score, (int, float))
    assert 0.0 <= score <= 1.0


def test_precision_metric_edge_cases():
    """Test PrecisionMetric with edge cases."""
    from pysad.evaluation import PrecisionMetric
    import warnings
    
    metric = PrecisionMetric()
    
    # Test with no positive predictions (binary predictions)
    metric.update(0, 0)  # True negative
    metric.update(1, 0)  # False negative
    metric.update(0, 0)  # True negative
    
    # Precision should handle division by zero gracefully
    # Suppress the sklearn warning since we're testing edge cases
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Precision is ill-defined and being set to 0.0")
        precision = metric.get()
        assert precision == 0.0


def test_precision_metric_normal_case():
    """Test PrecisionMetric with normal binary predictions."""
    from pysad.evaluation import PrecisionMetric
    
    metric = PrecisionMetric()
    
    # Test with mixed predictions
    metric.update(1, 1)  # True positive
    metric.update(0, 1)  # False positive
    metric.update(1, 1)  # True positive
    
    # Precision should be 2/3
    precision = metric.get()
    assert abs(precision - 2.0/3.0) < 1e-10


def test_recall_metric_edge_cases():
    """Test RecallMetric with edge cases."""
    from pysad.evaluation import RecallMetric
    import warnings
    
    metric = RecallMetric()
    
    # Test with no positive labels (binary predictions)
    metric.update(0, 1)  # False positive
    metric.update(0, 0)  # True negative
    metric.update(0, 1)  # False positive
    
    # Recall should handle division by zero gracefully 
    # Suppress the sklearn warning since we're testing edge cases
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Recall is ill-defined and being set to 0.0")
        recall = metric.get()
        assert recall == 0.0


def test_recall_metric_normal_case():
    """Test RecallMetric with normal binary predictions."""
    from pysad.evaluation import RecallMetric
    
    metric = RecallMetric()
    
    # Test with mixed predictions
    metric.update(1, 1)  # True positive
    metric.update(1, 0)  # False negative
    metric.update(1, 1)  # True positive
    
    # Recall should be 2/3
    recall = metric.get()
    assert abs(recall - 2.0/3.0) < 1e-10


def test_aupr_metric_comprehensive():
    """Test AUPRMetric with various scenarios."""
    from pysad.evaluation import AUPRMetric
    
    metric = AUPRMetric()
    
    # Test perfect prediction
    metric.update(1, 0.9)
    metric.update(0, 0.1)
    metric.update(1, 0.8)
    metric.update(0, 0.2)
    
    aupr = metric.get()
    assert 0.0 <= aupr <= 1.0


def test_auroc_metric_comprehensive():
    """Test AUROCMetric with various scenarios."""
    from pysad.evaluation import AUROCMetric
    import numpy as np
    
    metric = AUROCMetric()
    
    # Test random prediction (should be around 0.5)
    np.random.seed(42)
    for _ in range(20):
        y_true = np.random.randint(0, 2)
        y_pred = np.random.random()
        metric.update(y_true, y_pred)
    
    auroc = metric.get()
    assert 0.0 <= auroc <= 1.0


def test_auroc_single_class_error():
    """Test AUROCMetric with single class throws expected error."""
    from pysad.evaluation import AUROCMetric
    import pytest
    
    metric = AUROCMetric()
    
    # Test with all same labels (should throw ValueError)
    metric.update(1, 0.5)
    metric.update(1, 0.7)
    metric.update(1, 0.3)
    
    # AUROC should throw ValueError with single class
    with pytest.raises(ValueError, match="Only one class present"):
        metric.get()



import numpy as np
from pysad.core.base_model import BaseModel
from pysad.transform.projection.streamhash_projector import StreamhashProjector
from pysad.utils import get_minmax_array


class xStream(BaseModel):
    """The xStream model for row-streaming data :cite:`xstream`. It first projects the data via streamhash projection. It then fits half space chains by reference windowing. It scores the instances using the window fitted to the reference window.

    Args:
        n_components (int): The number of components for streamhash projection (Default=100).
        n_chains (int): The number of half-space chains (Default=100).
        depth (int): The maximum depth for the chains (Default=25).
        window_size (int): The size (and the sliding length) of the reference window (Default=25).
    """

    def __init__(
            self,
            num_components=100,
            n_chains=100,
            depth=25,
            window_size=25):
        self.streamhash = StreamhashProjector(num_components=num_components)
        deltamax = np.ones(num_components) * 0.5
        deltamax[np.abs(deltamax) <= 0.0001] = 1.0
        self.window_size = window_size
        self.hs_chains = _HSChains(
            deltamax=deltamax,
            n_chains=n_chains,
            depth=depth)

        self.step = 0
        self.cur_window = []
        self.ref_window = None

    def fit_partial(self, X, y=None):
        """Fits the model to next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to fit.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: Returns the self.
        """
        self.step += 1

        X = self.streamhash.fit_transform_partial(X)

        X = X.reshape(1, -1)
        self.cur_window.append(X)

        self.hs_chains.fit(X)

        if self.step % self.window_size == 0:
            self.ref_window = self.cur_window
            self.cur_window = []
            deltamax = self._compute_deltamax()
            self.hs_chains.set_deltamax(deltamax)
            self.hs_chains.next_window()

        return self

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            score (float): The anomalousness score of the input instance.
        """
        X = self.streamhash.fit_transform_partial(X)
        X = X.reshape(1, -1)
        score = self.hs_chains.score(X).flatten()

        return score

    def _compute_deltamax(self):
        # mx = np.max(np.concatenate(self.ref_window, axis=0), axis=0)
        # mn = np.min(np.concatenate(self.ref_window, axis=0), axis=0)
        mn, mx = get_minmax_array(np.concatenate(self.ref_window, axis=0))

        deltamax = (mx - mn) / 2.0
        deltamax[np.abs(deltamax) <= 0.0001] = 1.0

        return deltamax


class _Chain:

    def __init__(self, deltamax, depth):
        k = len(deltamax)

        self.depth = depth
        self.fs = [np.random.randint(0, k) for d in range(depth)]
        self.cmsketches = [{} for i in range(depth)] * depth
        self.cmsketches_cur = [{} for i in range(depth)] * depth

        self.deltamax = deltamax  # feature ranges
        self.rand_arr = np.random.rand(k)
        self.shift = self.rand_arr * deltamax

        self.is_first_window = True

    def fit(self, X):
        prebins = np.zeros(X.shape, dtype=np.float64)
        depthcount = np.zeros(len(self.deltamax), dtype=np.int32)
        for depth in range(self.depth):
            f = self.fs[depth]
            depthcount[f] += 1

            if depthcount[f] == 1:
                prebins[:, f] = (X[:, f] + self.shift[f]) / self.deltamax[f]
            else:
                prebins[:, f] = 2.0 * prebins[:, f] - \
                    self.shift[f] / self.deltamax[f]

            if self.is_first_window:
                cmsketch = self.cmsketches[depth]
                for prebin in prebins:
                    l_index = tuple(np.floor(prebin).astype(np.int32))
                    if l_index not in cmsketch:
                        cmsketch[l_index] = 0
                    cmsketch[l_index] += 1

                self.cmsketches[depth] = cmsketch

                self.cmsketches_cur[depth] = cmsketch

            else:
                cmsketch = self.cmsketches_cur[depth]

                for prebin in prebins:
                    l_index = tuple(np.floor(prebin).astype(np.int32))
                    if l_index not in cmsketch:
                        cmsketch[l_index] = 0
                    cmsketch[l_index] += 1

                self.cmsketches_cur[depth] = cmsketch

        return self

    def bincount(self, X):
        scores = np.zeros((X.shape[0], self.depth))
        prebins = np.zeros(X.shape, dtype=np.float64)
        depthcount = np.zeros(len(self.deltamax), dtype=np.int32)
        for depth in range(self.depth):
            f = self.fs[depth]
            depthcount[f] += 1

            if depthcount[f] == 1:
                prebins[:, f] = (X[:, f] + self.shift[f]) / self.deltamax[f]
            else:
                prebins[:, f] = 2.0 * prebins[:, f] - \
                    self.shift[f] / self.deltamax[f]

            cmsketch = self.cmsketches[depth]
            for i, prebin in enumerate(prebins):
                l_index = tuple(np.floor(prebin).astype(np.int32))
                if l_index not in cmsketch:
                    scores[i, depth] = 0.0
                else:
                    scores[i, depth] = cmsketch[l_index]

        return scores

    def score(self, X):
        # scale score logarithmically to avoid overflow:
        #    score = min_d [ log2(bincount x 2^d) = log2(bincount) + d ]
        scores = self.bincount(X)
        depths = np.array([d for d in range(1, self.depth + 1)])
        scores = np.log2(1.0 + scores) + depths  # add 1 to avoid log(0)
        return -np.min(scores, axis=1)

    def next_window(self):
        self.is_first_window = False
        self.cmsketches = self.cmsketches_cur
        self.cmsketches_cur = [{} for _ in range(self.depth)] * self.depth


class _HSChains:
    def __init__(self, deltamax, n_chains=100, depth=25):
        self.nchains = n_chains
        self.depth = depth
        self.chains = []

        for i in range(self.nchains):

            c = _Chain(deltamax=deltamax, depth=self.depth)
            self.chains.append(c)

    def score(self, X):
        scores = np.zeros(X.shape[0])
        for ch in self.chains:
            scores += ch.score(X)

        scores /= float(self.nchains)
        return scores

    def fit(self, X):
        for ch in self.chains:
            ch.fit(X)

    def next_window(self):
        for ch in self.chains:
            ch.next_window()

    def set_deltamax(self, deltamax):
        for ch in self.chains:
            ch.deltamax = deltamax
            ch.shift = ch.rand_arr * deltamax



from pysad.core.base_streamer import BaseStreamer
import numpy as np


class ArrayStreamer(BaseStreamer):
    """Simulator class to iterate array(s).

    Args:
        shuffle (bool): Whether shuffle the data initially (Default=False).
    """

    def __init__(self, shuffle=False):
        self.shuffle = shuffle

    def iter(self, X, y=None):
        """Iterates array of features and possibly labels.

        Args:
            X (np.array of shape (num_instances, num_features)): The features array.
            y (np.array of shape (num_instances, ): The array containing labels (Default=None).
        """
        indices = list(range(len(X)))
        if self.shuffle:
            np.random.shuffle(indices)

        if y is None:
            for i in indices:
                yield X[i]
        else:
            assert len(X) == len(y)
            for i in indices:
                yield X[i], y[i]



from pysad.core.base_model import BaseModel


class RobustRandomCutForest(BaseModel):
    """Robust Random Cut Forest model :cite:`guha2016robust`. The implementation uses `rrcf library <https://github.com/kLabUM/rrcf>`_ :cite:`bartos_2019_rrcf`.

        Args:
            num_trees (int): The number of trees.
            shingle_size (int): The shingle size (Default=4).
            tree_size (int): The tree size (Default=256).
    """

    def __init__(self, num_trees=4, shingle_size=4, tree_size=256):
        from rrcf import rrcf

        self.tree_size = tree_size
        self.shingle_size = shingle_size
        self.num_trees = num_trees

        self.forest = []
        for _ in range(self.num_trees):
            tree = rrcf.RCTree()
            self.forest.append(tree)

        self.index = 0

    def fit_partial(self, X, y=None):
        """Fits the model to next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to fit.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: Returns the self.
        """
        for tree in self.forest:
            if len(tree.leaves) > self.tree_size:
                tree.forget_point(self.index - self.tree_size)

            tree.insert_point(X, index=self.index)

        self.index += 1

        return self

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            float: The anomalousness score of the input instance.
        """

        score = 0.0
        for tree in self.forest:
            leaf = tree.find_duplicate(X)
            if leaf is None:
                tree.insert_point(X, index="test_point")
                score += 1.0 * tree.codisp("test_point") / self.num_trees
                tree.forget_point("test_point")
            else:
                score += 1.0 * tree.codisp(leaf) / self.num_trees

        return score



from abc import ABC, abstractmethod
from pysad.utils import _iterate
import numpy as np


class BaseModel(ABC):
    """Abstract base class for the models.
    """

    @abstractmethod
    def fit_partial(self, X, y=None):
        """Fits the model to next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to fit.
            y (int): The label of the instance (Optional for unsupervised models, default=None).

        Returns:
            object: Returns the self.
        """
        pass

    @abstractmethod
    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            float: The anomalousness score of the input instance.
        """
        pass

    def fit_score_partial(self, X, y=None):
        """Applies fit_partial and score_partial to the next instance, respectively.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to fit and score.
            y (int): The label of the instance (Optional for unsupervised models, default=None).

        Returns:
            float: The anomalousness score of the input instance.
        """
        return self.fit_partial(X, y).score_partial(X)

    def fit(self, X, y=None):
        """Fits the model to all instances in order.

        Args:
            X (np.float64 array of shape (num_instances, num_features)): The instances in order to fit.
            y (int): The labels of the instances in order to fit (Optional for unsupervised models, default=None).

        Returns:
            object: Fitted model.
        """
        for xi, yi in _iterate(X, y):
            self.fit_partial(xi, yi)

        return self

    def score(self, X):
        """Scores all instaces via score_partial iteratively.

        Args:
            X (np.float64 array of shape (num_instances, num_features)): The instances in order to score.

        Returns:
            np.float64 array of shape (num_instances,): The anomalousness scores of the instances in order.
        """
        y_pred = np.empty(X.shape[0], dtype=np.float64)
        for i, (xi, _) in enumerate(_iterate(X)):
            score = self.score_partial(xi)
            y_pred[i] = np.asarray(score).item() if np.asarray(score).ndim > 0 else score

        return y_pred

    def fit_score(self, X, y=None):
        """This helper method applies fit_score_partial to all instances in order.

        Args:
            X (np.float64 array of shape (num_instances, num_features)): The instances in order to fit.
            y (np.int32 array of shape (num_instances, )): The labels of the instances in order to fit (Optional for unsupervised models, default=None).

        Returns:
            np.float64 array of shape (num_instances,): The anomalousness scores of the instances in order.
        """
        y_pred = np.zeros(X.shape[0], dtype=np.float64)
        for i, (xi, yi) in enumerate(_iterate(X, y)):
            # Extract scalar value to avoid deprecation warning
            score = self.fit_score_partial(xi, yi)
            if hasattr(score, 'item'):
                score = score.item()
            y_pred[i] = score

        return y_pred



from .pyod_model import PYODModel


class OneFitModel(PYODModel):
    """The wrapper model fits the `model_cls` to the initial instnaces. Then it scores all incoming instances, with this fitted model.

        Args:
            model_cls (class):     The model class to be instantiated.
            initial_X (np.float64 array of shape (num_initial_instances, num_features)): Initial instances to fit.
            initial_y (np.int32 array of shape (num_initial_instances,): Initial window's ground truth labels. Used if not None. Needs to be `None` for the unsupervised `model_cls` models. (Default=None).
            **kwargs (Keyword arguments): Keyword arguments that are passed to the `model_cls`.
    """

    def __init__(self, model_cls, initial_X, initial_y=None, **kwargs):
        super().__init__(model_cls, **kwargs)

        self.initial_y = initial_y
        self.initial_X = initial_X
        self._fit_model()

    def fit_partial(self, X, y=None):
        """This method is ignored. Added for convenience.

        Args:
            X: any
            y: any

        Returns:
            object: Returns the self.
        """
        return self

    def _fit_model(self):

        self.reset_model()

        if self.initial_y is None:
            self.model.fit(self.initial_X)
        else:
            self.model.fit(self.initial_X, self.initial_y)

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (np.float64 array of shape (num_features,)): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances.

        Returns:
            float: The anomalousness score of the input instance.
        """
        score = self.model.decision_function([X])[0]

        return score



from setuptools import setup, find_packages
import os

with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.rst'), encoding='utf-8') as f:
    readme = f.read()

with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'requirements.txt'),
          encoding='utf-8') as f:
    requirements = f.read().splitlines()

ver_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'pysad', 'version.py')
with open(ver_file) as f:
    exec(f.read())
VERSION = __version__

setup(
    name='pysad',
    version=VERSION,
    url='https://github.com/selimfirat/pysad',
    license='3-Clause BSD',
    author='Selim Firat Yilmaz',
    author_email='yilmazselimfirat@gmail.com',
    description='PySAD is an open-source python framework for anomaly detection on streaming multivariate data. ',
    long_description=readme,
    long_description_content_type='text/x-rst',
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    install_requires=requirements,
    setup_requires=['setuptools>=38.6.0'],
    download_url="https://github.com/selimfirat/pysad/archive/master.zip",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Education',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Information Technology',
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development",
        "License :: OSI Approved :: BSD License",
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
    python_requires=">=3.10"
)



def test_reference_window(test_path):
    from sklearn.utils import shuffle
    from pysad.models.integrations import ReferenceWindowModel
    from pysad.utils import Data
    from pysad.evaluation import AUROCMetric
    from pysad.utils import ArrayStreamer
    import os
    from pyod.models.iforest import IForest

    data = Data(os.path.join(test_path,"../../examples/data"))

    X_all, y_all = data.get_data("arrhythmia.mat")
    X_all, y_all = shuffle(X_all, y_all)

    model = ReferenceWindowModel(model_cls=IForest, window_size=240, sliding_size=30,
                                 initial_window_X=X_all[:100])

    iterator = ArrayStreamer(shuffle=False)

    auroc = AUROCMetric()

    y_pred = []
    for X, y in iterator.iter(X_all[100:], y_all[100:]):
        model.fit_partial(X)
        score = model.score_partial(X)

        y_pred.append(score)

        auroc.update(y, score)

    print("AUROC: ", auroc.get())


def test_one_fit(test_path):
    from sklearn.utils import shuffle
    from pysad.utils import Data
    from pysad.evaluation import AUROCMetric
    from pysad.utils import ArrayStreamer
    import os
    from pyod.models.iforest import IForest
    from pysad.models.integrations.one_fit_model import OneFitModel

    data = Data(os.path.join(test_path, "../../examples/data"))

    X_all, y_all = data.get_data("arrhythmia.mat")
    print(X_all, y_all)
    X_all, y_all = shuffle(X_all, y_all)

    model = OneFitModel(model_cls=IForest, initial_X=X_all[:100])

    iterator = ArrayStreamer(shuffle=False)

    auroc = AUROCMetric()

    y_pred = []
    for X, y in iterator.iter(X_all[100:], y_all[100:]):
        model.fit_partial(X)
        score = model.score_partial(X)

        y_pred.append(score)

        auroc.update(y, score)

    print("AUROC: ", auroc.get())


def test_one_fit_model_basic():
    """Test OneFitModel basic functionality."""
    from pysad.models.integrations.one_fit_model import OneFitModel
    from pyod.models.iforest import IForest
    import numpy as np
    
    # Create test data
    np.random.seed(42)
    initial_X = np.random.random((50, 2))
    test_X = np.random.random((20, 2))
    
    # Test OneFitModel
    model = OneFitModel(model_cls=IForest, initial_X=initial_X)
    
    # Test that fit_partial does nothing but returns self
    result = model.fit_partial(test_X[0])
    assert result is model
    
    # Test scoring
    for i in range(len(test_X)):
        score = model.score_partial(test_X[i])
        assert isinstance(score, (int, float))


def test_one_fit_model_with_labels():
    """Test OneFitModel with supervised learning."""
    from pysad.models.integrations.one_fit_model import OneFitModel
    from pyod.models.iforest import IForest
    import numpy as np
    import warnings
    
    # Create test data with labels
    np.random.seed(42)
    initial_X = np.random.random((50, 2))
    
    # Test OneFitModel with labels
    model = OneFitModel(model_cls=IForest, initial_X=initial_X)
    
    # Test scoring
    test_instance = np.random.random(2)
    score = model.score_partial(test_instance)
    assert isinstance(score, (int, float))


def test_reference_window_model_basic():
    """Test ReferenceWindowModel basic functionality."""
    from pysad.models.integrations.reference_window_model import ReferenceWindowModel
    from pyod.models.iforest import IForest
    import numpy as np
    
    # Create test data
    np.random.seed(42)
    initial_X = np.random.random((30, 2))
    
    # Test ReferenceWindowModel
    model = ReferenceWindowModel(
        model_cls=IForest,
        window_size=20,
        sliding_size=5,
        initial_window_X=initial_X
    )
    
    # Test fitting and scoring
    test_instances = np.random.random((15, 2))
    
    for i, instance in enumerate(test_instances):
        model.fit_partial(instance)
        score = model.score_partial(instance)
        assert isinstance(score, (int, float))


def test_reference_window_model_without_initial():
    """Test ReferenceWindowModel without initial window."""
    from pysad.models.integrations.reference_window_model import ReferenceWindowModel
    from pyod.models.iforest import IForest
    import numpy as np
    
    # Test ReferenceWindowModel without initial window
    model = ReferenceWindowModel(
        model_cls=IForest,
        window_size=10,
        sliding_size=3
    )
    
    # Test fitting and scoring
    np.random.seed(42)
    test_instances = np.random.random((25, 2))
    
    for i, instance in enumerate(test_instances):
        model.fit_partial(instance)
        score = model.score_partial(instance)
        assert isinstance(score, (int, float))


def test_reference_window_model_with_labels():
    """Test ReferenceWindowModel with labels."""
    from pysad.models.integrations.reference_window_model import ReferenceWindowModel
    from pyod.models.iforest import IForest
    import numpy as np
    import warnings
    
    # Create test data with labels
    np.random.seed(42)
    initial_X = np.random.random((20, 2))
    initial_y = np.random.randint(0, 2, 20)
    
    # Test ReferenceWindowModel with labels
    # Suppress PyOD warning about labels in unsupervised learning
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="y should not be presented in unsupervised learning")
        model = ReferenceWindowModel(
            model_cls=IForest,
            window_size=15,
            sliding_size=5,
            initial_window_X=initial_X,
            initial_window_y=initial_y
        )
    
        # Test fitting and scoring
        test_instances = np.random.random((10, 2))
        test_labels = np.random.randint(0, 2, 10)
        
        for i, (instance, label) in enumerate(zip(test_instances, test_labels)):
            model.fit_partial(instance, label)
            score = model.score_partial(instance)
            assert isinstance(score, (int, float))


def test_reference_window_model_window_update():
    """Test ReferenceWindowModel window update mechanism."""
    from pysad.models.integrations.reference_window_model import ReferenceWindowModel
    from pyod.models.iforest import IForest
    import numpy as np
    
    # Test with small window and sliding size for easier testing
    np.random.seed(42)
    initial_X = np.random.random((5, 2))
    
    model = ReferenceWindowModel(
        model_cls=IForest,
        window_size=5,
        sliding_size=2,  # Model should retrain every 2 instances
        initial_window_X=initial_X
    )
    
    # Add instances and verify model retrains
    test_instances = np.random.random((8, 2))
    
    for i, instance in enumerate(test_instances):
        model.fit_partial(instance)
        score = model.score_partial(instance)
        assert isinstance(score, (int, float))
        
        # Check internal state
        assert len(model.cur_window_X) <= model.window_size




def test_calibrators():
    from pysad.transform.probability_calibration import GaussianTailProbabilityCalibrator
    import numpy as np
    from pysad.transform.probability_calibration import ConformalProbabilityCalibrator
    from pysad.utils import fix_seed
    fix_seed(61)

    scores = np.random.rand(100)

    calibrators = {
        GaussianTailProbabilityCalibrator: {},
        ConformalProbabilityCalibrator: {}
    }

    for calibrator_cls, args in calibrators.items():
        calibrator = calibrator_cls(**args)
        calibrated_scores = calibrator.fit_transform(scores)

        assert calibrated_scores.shape == scores.shape
        assert not np.isnan(calibrated_scores).any()

        calibrator = calibrator_cls(**args).fit(scores)
        assert type(calibrator) is calibrator_cls
        calibrated_scores = calibrator.fit_transform(scores)

        assert calibrated_scores.shape == scores.shape
        assert not np.isnan(calibrated_scores).any()




def test_ensemblers():
    import numpy as np
    from pysad.transform.ensemble import AverageScoreEnsembler, MaximumScoreEnsembler, MedianScoreEnsembler, \
    AverageOfMaximumScoreEnsembler, MaximumOfAverageScoreEnsembler

    scores = np.random.rand(100, 10)

    ensemblers = {
        AverageScoreEnsembler: {},
        MaximumScoreEnsembler: {},
        MedianScoreEnsembler: {},
        AverageOfMaximumScoreEnsembler: {},
        MaximumOfAverageScoreEnsembler: {}
    }

    for ensembler_cls, params_dict in ensemblers.items():
        ensembler = ensembler_cls(**params_dict)
        ensembled_scores = ensembler.fit_transform(scores)

        assert ensembled_scores.shape == (scores.shape[0], )

        ensembler = ensembler_cls(**params_dict).fit(scores)
        ensembled_scores = ensembler.transform(scores)

        assert ensembled_scores.shape == (scores.shape[0], )



from pysad.core.base_statistic import UnivariateStatistic
from pysad.statistics.count_meter import CountMeter
from pysad.statistics.sum_meter import SumMeter
from pysad.statistics.sum_squares_meter import SumSquaresMeter


class VarianceMeter(UnivariateStatistic):
    """The statistic that keeps track of the variance of the values. The variance formula is: (sum_squares - (sum**2)/count)/count.

    Attrs:
        sum_meter (pyod.statistics.SumMeter object): SumMeter object.
        sum_squares_meter (pyod.statistics.SumSquaresMeter object): SumSquaresMeter object.
        count_meter (pyod.statistics.CountMeter object): CountMeter object.
    """

    def __init__(self):
        self.sum_meter = SumMeter()
        self.sum_squares_meter = SumSquaresMeter()
        self.count_meter = CountMeter()

    def update(self, num):
        """Updates the statistic with the value for a timestep.

        Args:
            num (float): The incoming value, for which the statistic is used.

        Returns:
            object: self.

        """
        self.sum_squares_meter.update(num)
        self.count_meter.update(num)
        self.sum_meter.update(num)

        return self

    def remove(self, num):
        """Updates the statistic by removing particular value.

        Args:
            num (float): The value to be removed.

        Returns:
            object: self.

        """
        self.sum_squares_meter.remove(num)
        self.sum_meter.remove(num)
        self.count_meter.remove(num)

        return self

    def get(self):
        """ Method to obtain the tracked statistic.

        Returns:
            float: The statistic.
        """
        sum_squares = self.sum_squares_meter.get()
        sum = self.sum_meter.get()
        count = self.count_meter.get()

        var = (sum_squares - (sum**2) / count) / count

        return var



from pysad.core.base_model import BaseModel
from pysad.statistics.median_meter import MedianMeter


class MedianAbsoluteDeviation(BaseModel):
    """Median Absolute Deviation method :cite:`hochenbaum2017automatic`.

    Args:
        absolute (bool): Whether to output score's absolute value (Default=True).
        b (float): The default value `1.4826` is used for normally distributed data. See :cite:`hochenbaum2017automatic` for details. (Default=1.4826).
    """

    def __init__(self, absolute=True, b=1.4826):
        self.b = b
        self.absolute = absolute
        self.median_meter = MedianMeter()
        self.mad_meter = MedianMeter()

    def fit_partial(self, X, y=None):
        """Fits the model to next instance.

        Args:
            X (float): The instance to fit. Note that this model is univariate.
            y (int): Ignored since the model is unsupervised (Default=None).

        Returns:
            object: Returns the self.
        """
        assert len(X) == 1  # Only for time series

        self.median_meter.update(X)
        self.mad_meter.update(X)

        return self

    def score_partial(self, X):
        """Scores the anomalousness of the next instance.

        Args:
            X (float): The instance to score. Higher scores represent more anomalous instances whereas lower scores correspond to more normal instances. Note that this model is univariate.

        Returns:
            float: The anomalousness score of the input instance.
        """
        median = self.median_meter.get()
        mad = self.b * self.mad_meter.get()
        score = (X - median) / (mad + 1e-10)

        return abs(score) if self.absolute else score



import numpy as np
from . import dA as AE
from . import corClust as CC

# This class represents a KitNET machine learner.
# KitNET is a lightweight online anomaly detection algorithm based on an ensemble of autoencoders.
# For more information and citation, please see our NDSS'18 paper: Kitsune: An Ensemble of Autoencoders for Online Network Intrusion Detection
# For licensing information, see the end of this document

class KitNET:
    #n: the number of features in your input dataset (i.e., x \in R^n)
    #m: the maximum size of any autoencoder in the ensemble layer
    #AD_grace_period: the number of instances the network will learn from before producing anomaly scores
    #FM_grace_period: the number of instances which will be taken to learn the feature mapping. If 'None', then FM_grace_period=AM_grace_period
    #learning_rate: the default stochastic gradient descent learning rate for all autoencoders in the KitNET instance.
    #hidden_ratio: the default ratio of hidden to visible neurons. E.g., 0.75 will cause roughly a 25% compression in the hidden layer.
    #feature_map: One may optionally provide a feature map instead of learning one. The map must be a list,
    #           where the i-th entry contains a list of the feature indices to be assingned to the i-th autoencoder in the ensemble.
    #           For example, [[2,5,3],[4,0,1],[6,7]]
    def __init__(self,n,max_autoencoder_size=10,FM_grace_period=None,AD_grace_period=10000,learning_rate=0.1,hidden_ratio=0.75, feature_map = None):
        # Parameters:
        self.AD_grace_period = AD_grace_period
        if FM_grace_period is None:
            self.FM_grace_period = AD_grace_period
        else:
            self.FM_grace_period = FM_grace_period
        if max_autoencoder_size <= 0:
            self.m = 1
        else:
            self.m = max_autoencoder_size
        self.lr = learning_rate
        self.hr = hidden_ratio
        self.n = n

        # Variables
        self.n_trained = 0 # the number of training instances so far
        self.n_executed = 0 # the number of executed instances so far
        self.v = feature_map
        if self.v is None:
            print("Feature-Mapper: train-mode, Anomaly-Detector: off-mode")
        else:
            self.__createAD__()
            print("Feature-Mapper: execute-mode, Anomaly-Detector: train-mode")
        self.FM = CC.corClust(self.n) #incremental feature cluatering for the feature mapping process
        self.ensembleLayer = []
        self.outputLayer = None

    #If FM_grace_period+AM_grace_period has passed, then this function executes KitNET on x. Otherwise, this function learns from x.
    #x: a numpy array of length n
    #Note: KitNET automatically performs 0-1 normalization on all attributes.
    def process(self,x):
        if self.n_trained > self.FM_grace_period + self.AD_grace_period: #If both the FM and AD are in execute-mode
            return self.execute(x)
        else:
            self.train(x)
            return 0.0

    #force train KitNET on x
    #returns the anomaly score of x during training (do not use for alerting)
    def train(self,x):
        if self.n_trained <= self.FM_grace_period and self.v is None: #If the FM is in train-mode, and the user has not supplied a feature mapping
            #update the incremetnal correlation matrix
            self.FM.update(x)
            if self.n_trained == self.FM_grace_period: #If the feature mapping should be instantiated
                self.v = self.FM.cluster(self.m)
                self.__createAD__()
                print("The Feature-Mapper found a mapping: "+str(self.n)+" features to "+str(len(self.v))+" autoencoders.")
                print("Feature-Mapper: execute-mode, Anomaly-Detector: train-mode")
        else: #train
            ## Ensemble Layer
            S_l1 = np.zeros(len(self.ensembleLayer))
            for a in range(len(self.ensembleLayer)):
                # make sub instance for autoencoder 'a'
                xi = x[self.v[a]]
                S_l1[a] = self.ensembleLayer[a].train(xi)
            ## OutputLayer
            self.outputLayer.train(S_l1)
            if self.n_trained == self.AD_grace_period+self.FM_grace_period:
                print("Feature-Mapper: execute-mode, Anomaly-Detector: exeute-mode")
        self.n_trained += 1

    #force execute KitNET on x
    def execute(self,x):
        if self.v is None:
            raise RuntimeError('KitNET Cannot execute x, because a feature mapping has not yet been learned or provided. Try running process(x) instead.')
        else:
            self.n_executed += 1
            ## Ensemble Layer
            S_l1 = np.zeros(len(self.ensembleLayer))
            for a in range(len(self.ensembleLayer)):
                # make sub inst
                xi = x[self.v[a]]
                S_l1[a] = self.ensembleLayer[a].execute(xi)
            ## OutputLayer
            return self.outputLayer.execute(S_l1)

    def __createAD__(self):
        # construct ensemble layer
        for map in self.v:
            params = AE.dA_params(n_visible=len(map), n_hidden=0, lr=self.lr, corruption_level=0, gracePeriod=0, hiddenRatio=self.hr)
            self.ensembleLayer.append(AE.dA(params))

        # construct output layer
        params = AE.dA_params(len(self.v), n_hidden=0, lr=self.lr, corruption_level=0, gracePeriod=0, hiddenRatio=self.hr)
        self.outputLayer = AE.dA(params)

# Copyright (c) 2017 Yisroel Mirsky
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


"""
The :mod:`pysad.transform.ensemble` module consist of ensemblers to combine scores from multiple anomaly detectors.
"""
from .ensemblers import MaximumScoreEnsembler, AverageScoreEnsembler, MedianScoreEnsembler, AverageOfMaximumScoreEnsembler, MaximumOfAverageScoreEnsembler

__all__ = ["MaximumScoreEnsembler", "MedianScoreEnsembler", "AverageScoreEnsembler", "MaximumOfAverageScoreEnsembler", "AverageOfMaximumScoreEnsembler"]



"""
The :mod:`pysad.transform` module contains modules to preprocess inputs, calibrate probabilities, postprocess streaming scores and ensemble scores from multiple anomaly detectors.
"""

from . import ensemble, postprocessing, preprocessing, projection, probability_calibration

__all__ = ["ensemble", "postprocessing", "preprocessing", "projection", "probability_calibration"]



"""
The :mod:`pysad.transform.probability_calibration` module includes probability calibrators to convert module scores into true probabilities for decision-making on anomalousness.
"""
from .conformal_prediction import ConformalProbabilityCalibrator
from .gaussian_tail import GaussianTailProbabilityCalibrator

__all__ = ["ConformalProbabilityCalibrator", "GaussianTailProbabilityCalibrator"]



from pysad.core.base_model import BaseModel
import numpy as np


class RandomModel(BaseModel):
    """Random scorer that chooses a score between 0 and 1 ignoring the input.

    """

    def fit_partial(self, X, y=None):
        """This method is ignored. Added for convenience.

        Args:
            X: any
            y: any

        Returns:
            object: Returns the self.
        """
        return self

    def score_partial(self, X):
        """Randomly outputs a score from the uniform distribution.

        Args:
            X: any (Ignored)

        Returns:
            float: Uniform random between [0,1).
        """
        return np.random.uniform()



"""
The :mod:`pysad.evaluation` module includes evaluation metrics for anomaly detection on streaming data.
"""
from .windowed_metric import WindowedMetric
from .metrics import BaseSKLearnMetric, RecallMetric, PrecisionMetric, AUPRMetric, AUROCMetric

__all__ = ["BaseSKLearnMetric", "PrecisionMetric", "RecallMetric", "AUROCMetric", "AUPRMetric", "WindowedMetric"]



from abc import ABC, abstractmethod
import numpy as np
from pysad.utils import _iterate


class BasePostprocessor(ABC):
    """Base class for postprocessing methods.
    """

    @abstractmethod
    def fit_partial(self, score):
        """Fits particular (next) timestep's score to train the postprocessor.

        Args:
            score (float): Input score.

        Returns:
            object: self.
        """
        pass

    @abstractmethod
    def transform_partial(self, score):
        """Transforms given score.

        Args:
            score (float): Input score.

        Returns:
            float: Processed score.
        """
        pass

    def fit_transform_partial(self, score):
        """Shortcut method that iteratively applies fit_partial and transform_partial, respectively.

        Args:
            score (float): Input score.

        Returns:
            float: Processed score.
        """
        return self.fit_partial(score).transform_partial(score)

    def transform(self, scores):
        """Shortcut method that iteratively applies transform_partial to all instances in order.

        Args:
            np.float64 array of shape (num_instances,): Input scores.

        Returns:
            np.float64 array of shape (num_instances,): Processed scores.
        """
        processed_scores = np.empty(scores.shape[0], dtype=np.float64)
        for i, (score, _) in enumerate(_iterate(scores)):
            result = self.transform_partial(score)
            processed_scores[i] = np.asarray(result).item() if np.asarray(result).ndim > 0 else result

        return processed_scores

    def fit(self, scores):
        """Shortcut method that iteratively applies fit_partial to all instances in order.

        Args:
            np.float64 array of shape (num_instances,): Input scores.

        Returns:
            object: self.
        """
        for i, (score, _) in enumerate(_iterate(scores)):
            self.fit_partial(score)

        return self

    def fit_transform(self, scores):
        """Shortcut method that iteratively applies fit_transform_partial to all instances in order.

        Args:
            np.float64 array of shape (num_instances,): Input scores.

        Returns:
            np.float64 array of shape (num_instances,): Processed scores.
        """
        processed_scores = np.empty(scores.shape[0], dtype=np.float64)
        for i, (score, _) in enumerate(_iterate(scores)):
            result = self.fit_transform_partial(score)
            processed_scores[i] = np.asarray(result).item() if np.asarray(result).ndim > 0 else result

        return processed_scores
